package org.expeditee.auth;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.UnrecoverableEntryException;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.sql.SQLException;
import java.util.Base64;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;

import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import org.expeditee.auth.Mail.MailEntry;
import org.expeditee.auth.gui.MailBay;
import org.expeditee.core.Colour;
import org.expeditee.gui.DisplayController;
import org.expeditee.gui.Frame;
import org.expeditee.gui.FrameIO;
import org.expeditee.gui.MessageBay;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.settings.UserSettings;
import org.ngikm.cryptography.CryptographyConstants;

public class Actions implements CryptographyConstants {
	
	/**
	 * Display Expeditee Mail
	 * @throws IOException 
	 * @throws SQLException 
	 * @throws ClassNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 */
	public static void MailMode() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, SQLException, IOException {
		if (!DisplayController.isMailMode()) {
			Mail.clear();
			Authenticator.getInstance().loadMailDatabase();
		}
		DisplayController.ToggleMailMode();
	}
 	
	/**
	 * Action used to navigate the authorised user back to their desktop.
	 */
	public static void AuthGoToDesktop() {
		DisplayController.setCurrentFrame(FrameIO.LoadFrame(UserSettings.HomeFrame.get()), true);
	}
	
	/**
	 * Log out of current user and navigate to authentication1
	 */
	public static void AuthLogout() {
		UserSettings.Authenticated.set(false);
		setUser(System.getProperty("user.name"));
		DisplayController.setCurrentFrame(FrameIO.LoadFrame("authentication1"), true);
	}
	
	/**
	 * Action used to created a new user account.
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * @throws Exception 
	 */
	public static void AuthCreateAccount() throws Exception {
		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		final Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Username, AuthenticationTag.Password, AuthenticationTag.PasswordAgain);
		if (userdata.isPresent()) { 
			final Map<AuthenticationTag, String> userData = userdata.get();
			final String password = userData.get(AuthenticationTag.Password);
			final String passwordAgain = userData.get(AuthenticationTag.PasswordAgain);
			if (password.compareTo(passwordAgain) == 0) {
				createAccount(userData);
			} else {
				MessageBay.errorMessage("The passwords provided do not match.");
			}
		} else {
			MessageBay.errorMessage("Please enter a username and password ( + repeated password) to create an account."); 
		}
	}
	
	/**
	 * Action used to start authentication as a specified user.  
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * @throws Exception 
	 */
	public static void AuthLogin() throws Exception {
		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		final Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Username, AuthenticationTag.Password);
		if (userdata.isPresent()) { 
			login(userdata.get());
		}
		else {
			MessageBay.errorMessage("Please enter a username and password to log in.");
		}
	}
	
	/**
	 * Action used to change the currently authenticated users password.
	 * Attempts to use content from text items on frame, will default to java properties if they cannot be found.
	 * Will fail if it cannot find content from text items on frame and all required java properties are not present.
	 * Will fail if no user is currently logged in.
	 * @throws IOException 
	 * @throws CertificateException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 * @throws NoSuchAlgorithmException 
	 * @throws SQLException 
	 * @throws ClassNotFoundException 
	 */
	public static void AuthChangePassword() throws NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, IOException, ClassNotFoundException, SQLException {
		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		
		if (!UserSettings.Authenticated.get()) {
			MessageBay.errorMessage("You must be logged in to change your password.");
		} else {
			final Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Password, AuthenticationTag.NewPassword, AuthenticationTag.NewPasswordAgain);
			if (userdata.isPresent()) {
				final Map<AuthenticationTag, String> userData = userdata.get();
				if (userData.get(AuthenticationTag.NewPassword).compareTo(userData.get(AuthenticationTag.NewPasswordAgain)) != 0) {
					MessageBay.errorMessage("The passwords provided do not match.");
				} else {
					userData.put(AuthenticationTag.Username, UserSettings.UserName.get());
					changePassword(userData);
				}
			} else {
				MessageBay.errorMessage("Insufficient information provided to complete this action.");
			}
		}
	}
	
	/**
	 * Part of the process to secure an authorised account, the AuthAssociateEmail action associates a supplied email
	 * with the currently logged in user.  
	 */
	public static void AuthAssociateEmail() {
		Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
		
		if (!UserSettings.Authenticated.get()) {
			MessageBay.errorMessage("You must be logged in to perform this action.");
		} else { 
			Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.Email);
			if (userdata.isPresent()) {
				Map<AuthenticationTag, String> userData = userdata.get();
				userData.put(AuthenticationTag.Username, UserSettings.UserName.get());
				String emailRegex = "^[A-Z0-9._%+-]+@[A-Z0-9.-]+\\.[A-Z]{2,}$";
				if (!userData.get(AuthenticationTag.Email).toUpperCase().matches(emailRegex)) {
					MessageBay.errorMessage("Invalid email address supplied.");
				} else {
					try {
						associateEmail(userData);
						DisplayController.setCurrentFrame(FrameIO.LoadFrame("authentication5"), true);
					} catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | ClassNotFoundException | IOException | SQLException  e) {
						MessageBay.errorMessage("An unknown error has occured. Details sent to standard error.");
						e.printStackTrace();
					}
				}
			} else {
				MessageBay.errorMessage("Insufficient information provided to complete this action.");
			}
		}
	}
	
	/**
	 * Navigation action for progressing the process of recruiting colleagues to assist in password recovery.
	 * Hides certain content that AuthSubmitPWCollegues goes onto show if it does not fail.
	 */
	public static void AuthGotoColleagueSubmissionFrame() {
		Frame destination = FrameIO.LoadFrame("authentication7");
		DisplayController.setCurrentFrame(destination, true);
		Collection<Item> toHide = getByData(destination, "ShowOnProgress");
		for (Item i: toHide) {
			i.setVisible(false);
		}
	}
	
	/**
	 * Action used to start the process of formalising the password recovery process.
	 * @throws SQLException 
	 * @throws IOException 
	 * @throws ClassNotFoundException 
	 * @throws CertificateException 
	 * @throws NoSuchAlgorithmException 
	 * @throws FileNotFoundException 
	 * @throws KeyStoreException 
	 * @throws InvalidKeySpecException 
	 */
	public static void AuthSubmitPWColleagues() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, IOException, SQLException, InvalidKeySpecException {
		final Frame currentFrame = DisplayController.getCurrentFrame();
		Collection<Text> textItems = currentFrame.getTextItems();
		
		if (!UserSettings.Authenticated.get()) {
			MessageBay.errorMessage("You must be logged in to perform this action.");
		} else {
			Optional<Map<AuthenticationTag, String>> userdata = AuthenticationTag.fetchUserData(textItems, false, AuthenticationTag.ColleagueOne, AuthenticationTag.ColleagueTwo);
			if (userdata.isPresent()) {
				Map<AuthenticationTag, String> userData = userdata.get();
				userData.put(AuthenticationTag.Username, UserSettings.UserName.get());
				boolean succsess = submitPWColleagues(userData);
				if (succsess) {
					Collection<Item> toShow = getByData(currentFrame, "ShowOnProgress");
					for (Item i: toShow) {
						i.setVisible(true);
					}
					currentFrame.change();
				}
				FrameIO.SaveFrame(currentFrame);
				DisplayController.requestRefresh(false);
			} else {
				MessageBay.errorMessage("Insufficient information provided to complete this action.");
			}
		}
	}
	
	public static void AuthSetupPasswordRecovery() throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, IOException, SQLException, UnrecoverableEntryException {
		if (!UserSettings.Authenticated.get()) {
			MessageBay.errorMessage("You must be logged in to perform this action.");
		} else if (!Authenticator.getInstance().hasRegisteredEmail(UserSettings.UserName.get())) {
			Frame registerEmailFrame = FrameIO.LoadFrame("authentication4");
			DisplayController.setCurrentFrame(registerEmailFrame, true);
		} else if (!Authenticator.getInstance().hasRequestedColleagues(UserSettings.UserName.get()) && Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) == null) {
			Frame submitColleaguesFrame = FrameIO.LoadFrame("authentication5");
			DisplayController.setCurrentFrame(submitColleaguesFrame, true);
		} else if (Authenticator.getInstance().hasRequestedColleagues(UserSettings.UserName.get()) && Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) == null) {
			MessageBay.displayMessage("You have already nominated two colleagues to assist you in the process of password recovery and are awaiting their response."
					+ "  You will be alerted on Expeditee startup when they have both responded.");
		} else if (Authenticator.getInstance().getColleagues(UserSettings.UserName.get()) != null) {
			MessageBay.displayMessage("You have completed the Password Recovery Setup process, there is nothing more to do here.");
		}
	}
	
	public static void AuthConfirmPasswordColleagueRelationship(String colleagueName) {
		
	}
	
	public static void AuthDenyPasswordColleagueRelationship(String colleagueName) throws InvalidKeySpecException, NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		denyPasswordColleagueRelationship(colleagueName);
	}
	
	public static void AuthClearPWColleaguesNominated() {
		
	}
	
	private static void denyPasswordColleagueRelationship(String colleagueName) throws InvalidKeySpecException, NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, ClassNotFoundException, IOException, SQLException {
		String sender = UserSettings.UserName.get();
		String message = "You have received a reply from " + sender + " reguarding your request for assistance.";
		String message2 = "Unfortunately " + sender + " has indicated that they are unable to help you with your potential password recovery.";
		Map<String, String> options = new HashMap<String, String>();
		options.put("Clear Preview Colleague Nominations", "AuthClearPWColleaguesNominated");
		MailEntry mail = new MailEntry(sender, colleagueName, message, message2, options);
		Mail.sendMail(mail, Authenticator.getInstance().getPublicKey(colleagueName));
	}
	
	private static boolean submitPWColleagues(Map<AuthenticationTag, String> userData) throws KeyStoreException, FileNotFoundException, NoSuchAlgorithmException, CertificateException, ClassNotFoundException, IOException, SQLException, InvalidKeySpecException {
		// Confirm both colleagues exist
		String colleagueOne = userData.get(AuthenticationTag.ColleagueOne);
		String colleagueTwo = userData.get(AuthenticationTag.ColleagueTwo);
		PublicKey colleagueOneKey = null;
		PublicKey colleagueTwoKey = null;
		if (!FrameIO.canAccessFrameset(colleagueOne) && FrameIO.canAccessFrameset(colleagueTwo)) {
			MessageBay.errorMessage("At least one of the colleagues you specified does not have a Expeditee profile.");
			return false;
		} else if ((colleagueOneKey = Authenticator.getInstance().getPublicKey(colleagueOne)) == null || 
				(colleagueTwoKey = Authenticator.getInstance().getPublicKey(colleagueTwo)) == null) {
			MessageBay.errorMessage("At least one of the colleagues you specified is not set up to be authorised.");
			return false;
		} else {
			String sender = UserSettings.UserName.get();
			String message = "You have received a request for cooperation from your colleague " + sender;
			String message2 = "\"I would like to receive your help if I happen to forget my password.\"";
			Map<String, String> arguments = new HashMap<String, String>();
			arguments.put("I Agree To Assist " + sender + " If They Forget Their Password" , "AuthConfirmPasswordColleagueRelationship " + sender);
			arguments.put("I Wish To Excuse Myself From This Responsibility", "AuthDenyPasswordColleagueRelationship " + sender);
			MailEntry mail = new MailEntry(sender, colleagueOne, message, message2, arguments);
			Mail.sendMail(mail, colleagueOneKey);
			mail = new MailEntry(sender, colleagueTwo, message, message2, arguments);
			Mail.sendMail(mail, colleagueTwoKey);
			Authenticator.getInstance().markRequestedColleagues(UserSettings.UserName.get());
			return true;
		}
	}

	/*
	 * Function used to authenticate as a specified user (via function arguments).
	 */
	private static void login(final Map<AuthenticationTag, String> userdata) throws Exception {
		final String username = userdata.get(AuthenticationTag.Username);
		final String password = userdata.get(AuthenticationTag.Password);
		
		final SecretKey personalKey = Authenticator.getInstance().getSecretKey(username, password);
		if (personalKey == null) {
			// Incorrect username and password
			MessageBay.errorMessage("The username + password combination was incorrect.");
		} else {
			// Proceed with login process.
			org.expeditee.settings.auth.secrets.KeyList.PersonalKey.setSetting(new Text(Base64.getEncoder().encodeToString(personalKey.getEncoded())));
			setUser(username);
			MessageBay.displayMessage("Logged in as: " + UserSettings.UserName.get());
			
			// Set Public Key and private key in settings datastructure
			String profileName = UserSettings.ProfileName.get();
			int lastNumber = FrameIO.getLastNumber(profileName);
			for (int i = 1; i <= lastNumber; i++) {
				Frame frame = FrameIO.LoadFrame(profileName + i);
				if (frame == null) {
					continue;
				}
				Collection<Item> items = frame.getAllItems();
				boolean foundPublic = false;
				boolean foundPrivate = false;
				for (Item item: items) {
					if (item.getText().compareTo("PublicKey") == 0) {
						org.expeditee.settings.auth.KeyList.PublicKey.set((Text) item);
						foundPublic = true;
					} else if (item.getText().compareTo("PrivateKey") == 0) {
						org.expeditee.settings.auth.secrets.KeyList.PrivateKey.set((Text) item);
						foundPrivate = true;
					}
				}
				if (foundPublic && foundPrivate) { break; }
			}
						
			// Check mail
			MailBay.clear();
			Authenticator.getInstance().loadMailDatabase();
			Text keyItem = org.expeditee.settings.auth.secrets.KeyList.PrivateKey.get();
			if (keyItem.getData() != null) {
				String keyEncoded = keyItem.getData().get(0);
				byte[] keyBytes = Base64.getDecoder().decode(keyEncoded);
				PrivateKey key = KeyFactory.getInstance(AsymmetricAlgorithm).generatePrivate(new PKCS8EncodedKeySpec(keyBytes));
				List<MailEntry> mailForLoggingInUser = Mail.getEntries(UserSettings.UserName.get(), key);
				for (MailEntry mail: mailForLoggingInUser) {
					MailBay.addMessage(mail.message, mail.message2, mail.options);
				}
			}
			
			UserSettings.Authenticated.set(true);
			// navigate to logged in profile
			DisplayController.setCurrentFrame(loadProfile(username), false);
		}
	}
	
	/*
	 * Creates an account from specified user data.
	 */
	private static void createAccount(final Map<AuthenticationTag, String> userdata) throws Exception {
		String username = userdata.get(AuthenticationTag.Username);
		String password = userdata.get(AuthenticationTag.Password);
		if (FrameIO.LoadProfile(username) != null) {
			MessageBay.errorMessage("A Expeditee profile with this username already exists, please choose another.");
		} else {				
			// generate the newly created user a profile frame using the specialized default frame as a blueprint
			String defaultBackup = UserSettings.DEFAULT_PROFILE_NAME;
			UserSettings.DEFAULT_PROFILE_NAME = "multiusermodedefault";
			Frame profile = FrameIO.CreateNewProfile(username);
			UserSettings.DEFAULT_PROFILE_NAME = defaultBackup;
			
			// perform necessary changes to generated frame: remove default title, set background color, set custom title 
			profile.getTitleItem().delete();
			profile.setBackgroundColor(new Colour(1, 1, 0.39f));
			getByData(profile, "txtUsername").stream().forEach(i -> i.setText(username));
			getByText(profile, "“If Aphroditē be the godess of love").stream().forEach(i -> i.setWidth(300));
			getByText(profile, "Sleek like Aphroditē, Wise like Athēnâ.").stream().forEach(i -> i.setWidth(300));
			
			// update ownership
			int lastNumber = FrameIO.getLastNumber(profile.getFramesetName());
			for (int i = 1; i <= lastNumber; i++) {
				Frame frame = FrameIO.LoadFrame(profile.getFramesetName() + i);
				frame.setOwner(username);
				frame.setEncryptionLabel("Profile");
				Collection<Item> items = frame.getAllItems();
				for (Item item: items) {
					String newText = item.getText().replace("authadmin", username);
					item.setText(newText);
				}
			}
			
			// generate and store personal, public and private keys
			String backupUser = UserSettings.UserName.get();
			setUser(username);
			UserSettings.Authenticated.set(true);
			
			// personal key
			Random rand = new Random();
			byte[] keyBytes = new byte[16];
			rand.nextBytes(keyBytes);
			SecretKey key = new SecretKeySpec(keyBytes, SymmetricAlgorithm);
			Authenticator.getInstance().putKey(username, password, key);
			org.expeditee.settings.auth.secrets.KeyList.PersonalKey.setSetting(new Text(Base64.getEncoder().encodeToString(key.getEncoded())));
			
			// public+private key pair
			KeyPairGenerator keyGen = KeyPairGenerator.getInstance(AsymmetricAlgorithm);
			keyGen.initialize(1024);
			KeyPair keyPair = keyGen.generateKeyPair();
			
			org.expeditee.settings.auth.secrets.KeyList.PrivateKey.setSetting(new Text(Base64.getEncoder().encodeToString(keyPair.getPrivate().getEncoded())));
			org.expeditee.settings.auth.KeyList.PublicKey.setSetting(new Text(Base64.getEncoder().encodeToString(keyPair.getPublic().getEncoded())));
			
			// ensure save of keys
			lastNumber = FrameIO.getLastNumber(profile.getFramesetName());
			for (int i = 1; i <= lastNumber; i++) {
				Frame frame = FrameIO.LoadFrame(profile.getFramesetName() + i);
				frame.setChanged(true);
				FrameIO.SaveFrame(frame);
			}
			setUser(backupUser);
			UserSettings.Authenticated.set(false);
			
			// proceed to log into new account.
			login(userdata);
		}
	}
	
	/*
	 * Changes the recorded password for a user in the key store.
	 */
	private static void changePassword(final Map<AuthenticationTag, String> userdata) throws NoSuchAlgorithmException, KeyStoreException, FileNotFoundException, CertificateException, IOException, ClassNotFoundException, SQLException {
		final String username = userdata.get(AuthenticationTag.Username);
		final String password = userdata.get(AuthenticationTag.Password);
		final String newpassword = userdata.get(AuthenticationTag.NewPassword);

		final SecretKey key = Authenticator.getInstance().getSecretKey(username, password);
		if (key == null) {
			MessageBay.errorMessage("The username + existing password combination was incorrect.");
		} else {
			Authenticator.getInstance().putKey(username, newpassword, key);
			MessageBay.displayMessage("Password changed successfully.");
		}
	}
	
	private static void associateEmail(Map<AuthenticationTag, String> userdata) throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, ClassNotFoundException, IOException, SQLException {
		Authenticator.getInstance().putEmail(
				userdata.get(AuthenticationTag.Username), 
				userdata.get(AuthenticationTag.Email));
	}

	private static void setUser(final String username) {
		UserSettings.UserName.set(username);
		UserSettings.ProfileName.set(username);
		UserSettings.HomeFrame.set(username + "1");
		UserSettings.UserName.setDefault(username);
		UserSettings.ProfileName.setDefault(username);
		UserSettings.HomeFrame.setDefault(username + "1");
	}
	
	private static Frame loadProfile(final String username) throws Exception {
		final Frame frameOne = FrameIO.LoadProfile(username);
		if (frameOne != null) { 
			return frameOne;
		} else {
			final Frame profile = FrameIO.CreateNewProfile(username);
			for (int i = 0; i < FrameIO.getLastNumber(profile.getFramesetName()); i++) {
				final Frame frame = FrameIO.LoadFrame(profile.getFramesetName() + i);
				final Collection<Text> textItems = frame.getTextItems();
				for (final Text t: textItems) {
					if (t.getText().startsWith("User name:")) {
						t.setText("User name: " + UserSettings.UserName.get());
					} else if (t.getText().startsWith("Profile name:")) {
						t.setText("Profile name: " + UserSettings.ProfileName.get());
					}
				}
				FrameIO.ForceSaveFrame(frame);
			}
			return profile;
		}
	}

//	public static void AuthRecoverPasswordStage1() throws KeyStoreException, NoSuchAlgorithmException, CertificateException, FileNotFoundException, IOException {
//		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
//		final Optional<Text> txtUsername = textItems.stream().filter(t -> t.getData() != null && t.getData().contains("txtUsername")).findFirst();
//		final Optional<Text> txtRecoveryEmail = textItems.stream().filter(t -> t.getData() != null && t.getData().contains("txtRecoveryEmail")).findFirst();
//		final String username = txtUsername.isPresent() ? txtUsername.get().getText() : System.getProperty("user.name");
//		final String email = txtRecoveryEmail.isPresent() ? txtRecoveryEmail.get().getText() : System.getProperty("user.recoveryemail");
//		if (areAllPresent(new String[] { username, email })) {
//			// generate intergalactic number
//			final String intergalacticNumber = Authenticator.getInstance().newIntergalacticNumber(username, email);
//
//			// establish properties
//			final String from = "noreply@expeditee.com";
//			final Properties properties = System.getProperties();
//			
//			properties.setProperty("mail.transport.protocol", "smtp");
//			properties.setProperty("mail.smtp.host", "smtp.gmail.com");
//			properties.setProperty("mail.smtp.port", "465");
//			properties.setProperty("mail.smtp.starttls.enable", "true");
//			properties.setProperty("mail.smtp.auth", "true");
//			properties.setProperty("mail.smtp.debug", "true");
//			properties.setProperty("mail.smtp.auth", "true");
//			properties.setProperty("mail.smtp.socketFactory.port", "465");
//			properties.setProperty("mail.smtp.socketFactory.class", "javax.net.ssl.SSLSocketFactory");
//			properties.setProperty("mail.smtp.socketFactory.fallback", "false");
//			
//			final Session session = Session.getDefaultInstance(properties, new javax.mail.Authenticator() {
//				@Override
//				protected PasswordAuthentication getPasswordAuthentication() {
//					return new PasswordAuthentication("noreply.expeditee", "intergalacticnumber");
//				};
//			});
//
//			try {
//				// construct email message
//				final MimeMessage message = new MimeMessage(session);
//				message.setFrom(new InternetAddress(from));
//				message.addRecipient(Message.RecipientType.TO, new InternetAddress(email));
//				message.setSubject("Expeditee Password Recovery");
//				message.setText(intergalacticNumber);
//				
//				// send email message
//				Transport.send(message);
//			} catch (final MessagingException e) {
//				e.printStackTrace();
//			}
//			
//			FrameUtils.DisplayFrame("4", false, true);
//		} else {
//			MessageBay.errorMessage("To recover a lost password please fill in all the provided fields.");
//		}
//	}
//	
//	public static void AuthRecoverPasswordStage2() throws NoSuchAlgorithmException, KeyStoreException, CertificateException, FileNotFoundException, IOException {
//		final Collection<Text> textItems = DisplayController.getCurrentFrame().getTextItems();
//		final Optional<Text> txtUsername = textItems.stream().filter(t -> t.getData() != null && t.getData().contains("txtUsername")).findFirst();
//		final Optional<Text> txtRecoveryEmail = textItems.stream().filter(t -> t.getData() != null && t.getData().contains("txtRecoveryEmail")).findFirst();
//		final Optional<Text> txtIntergalacticNumber = textItems.stream().filter(t -> t.getData() != null && t.getData().contains("txtIntergalacticNumber")).findFirst();
//		final String username = txtUsername.isPresent() ? txtUsername.get().getText() : System.getProperty("user.name");
//		final String email = txtRecoveryEmail.isPresent() ? txtRecoveryEmail.get().getText() : System.getProperty("user.recoveryemail");
//		final String intergalacticNumber = txtIntergalacticNumber.isPresent() ? txtIntergalacticNumber.get().getText() : System.getProperty("user.intergalacticnumber");
//		if (areAllPresent(new String[] { username, email, intergalacticNumber })) {
//			// confirm intergalactic number matches what is stored
//			final boolean intergalaticNumberMatches = Authenticator.getInstance().confirmIntergalaticNumber(username, email, intergalacticNumber);
//			if (intergalaticNumberMatches) { 
//				final Frame changePasswordFrame = FrameIO.LoadFrame("5");
//				final Optional<Text> actionNewPassword = changePasswordFrame.getTextItems().stream().filter(t -> t.getAction().contains("AuthNewPassword")).findFirst();
//				if (actionNewPassword.isPresent()) {
//					actionNewPassword.get().setData(intergalacticNumber);
//				}
//				FrameUtils.DisplayFrame("5", false, true);
//			}
//		} else {
//			MessageBay.errorMessage("To recover a lost password please fill in all the provided fields.");
//		}
//	}

	public static void TickBox(final Text item) {
		if (item.getBackgroundColor() != Colour.RED) {
			item.setBackgroundColor(Colour.RED);
		} else {
			item.setBackgroundColor(Colour.GREEN);
		}
	}
	
	/*
	 *  Gets all items on a specified frame that contain the specified text.
	 */
	private static Collection<Item> getByText(final Frame frame, final String text) {
		final Collection<Item> allItems = frame.getAllItems();
		allItems.removeIf(i -> i.getText() == null || !i.getText().contains(text));
		return allItems;
	}
	
	/*
	 * Gets all items on a specified frame that contain the specified data.
	 */
	private static Collection<Item> getByData(final Frame frame, final String data) {
		final Collection<Item> allItems = frame.getAllItems();
		allItems.removeIf(i -> i.getData() == null || !i.hasData(data));
		return allItems;
	}
}
